import numpy as np
import pandas as pd
from fibgrid.realization import FibGrid
from pynetcf.time_series import GriddedNcContiguousRaggedTs


class AscatData(GriddedNcContiguousRaggedTs):
    """
    Class reading ASCAT SSM 6.25 km data.
    """

    def __init__(self, path, read_bulk=True):
        """
        Initialize ASCAT data.

        Parameters
        ----------
        path : str
            Path to dataset.
        read_bulk : bool, optional
            If "True" all data will be read in memory, if "False"
            only a single time series is read (default: False).
            Use "True" to process multiple GPIs in a loop and "False" to
            read/analyze a single time series.
        """
        grid = FibGrid(6.25)
        ioclass_kws = dict(read_bulk=read_bulk)
        super().__init__(path, grid, ioclass_kws=ioclass_kws)


def read_grid_point(loc, ascat_sm_path, read_bulk=False):
    """
    Read grid point for given lon/lat coordinates or grid_point.

    Parameters
    ----------
    loc : int, tuple
        Tuple is interpreted as longitude, latitude coordinate.
        Integer is interpreted as grid point index.
    ascat_sm_path : str
        Path to ASCAT soil moisture data.
    read_bulk : bool, optional
        If "True" all data will be read in memory, if "False"
        only a single time series is read (default: False).
        Use "True" to process multiple GPIs in a loop and "False" to
        read/analyze a single time series.
    """
    data = {}

    # print(f"Reading ASCAT soil moisture: {ascat_sm_path}")
    ascat_obj = AscatData(ascat_sm_path, read_bulk)

    if isinstance(loc, tuple):
        lon, lat = loc
        ascat_gpi, distance = ascat_obj.grid.find_nearest_gpi(lon, lat)
        print(f"ASCAT GPI: {ascat_gpi} - distance: {distance:8.3f} m")
    else:
        ascat_gpi = loc
        lon, lat = ascat_obj.grid.gpi2lonlat(ascat_gpi)
        # print(f"ASCAT GPI: {ascat_gpi}")

    ascat_ts = ascat_obj.read(ascat_gpi)

    if ascat_ts is None:
        raise RuntimeError(f"ASCAT soil moisture data not found: {ascat_sm_path}")

    # set observations to NaN with less then two observations
    valid = ascat_ts["num_sigma"] >= 2
    ascat_ts.loc[~valid, ["sm", "sigma40", "slope40", "curvature40"]] = np.nan
    data["ascat_ts"] = ascat_ts
    data["ascat_gpi"] = ascat_gpi
    data["ascat_lon"] = lon
    data["ascat_lat"] = lat

    return data


def extract_obs_ts(loc, ascat_path, obs_type="sm", read_bulk=False):
    """
    Read time series of given observation type.

    Parameters
    ----------
    loc : int, tuple
        Tuple is interpreted as longitude, latitude coordinate.
        Integer is interpreted as grid point index.
    ascat_path : str
        Path to ASCAT soil moisture data.
    obs : str, optional
        Observation type (default: "sm").
    read_bulk : bool, optional
        If "True" all data will be read in memory, if "False"
        only a single time series is read (default: False).
        Use "True" to process multiple GPIs in a loop and "False" to
        read/analyze a single time series.
    """
    data = read_grid_point(loc, ascat_path, read_bulk)
    ascat_ts = data.get("ascat_ts")
    # lat = data.get("ascat_lat")
    # lon = data.get("ascat_lon")
    # gpi = data.get("ascat_gpi")
    ts = ascat_ts.get(obs_type)
    ts.dropna(inplace=True)

    # return {"ts": pd.DataFrame(ts), "lon": lon, "lat": lat, "gpi": gpi}

    return pd.DataFrame(ts)



import eomaps
from smadi.metadata import indicators_thresholds


def set_thresholds(method):
    """
    Set the thresholds for the specified method based on the method name.

    parameters:
    -----------

    method: str
        The method name for which the thresholds are to be set. Supported methods are:
        'zscore', 'smapi', 'smdi', 'smca', 'smad', 'smci', 'smds', 'essmi', 'beta', 'gamma'
    """

    if method in indicators_thresholds.keys():
        return indicators_thresholds[method]
    else:
        return None


def set_extent(df, x="lon", y="lat", buffer=2):
    """
    Set the extent for the map based on the provided dataframe and buffer.

    parameters:
    -----------

    df: pd.DataFrame
        The dataframe containing the data.

    x: str
        The column name for the x-axis.

    y: str
        The column name for the y-axis.

    buffer: int
        The buffer to be added to the min and max values of the x and y axis.
    """

    min_lat = df[y].min() - buffer
    max_lat = df[y].max() + buffer
    min_lon = df[x].min() - buffer
    max_lon = df[x].max() + buffer

    return min_lon, max_lon, min_lat, max_lat


def set_bins(colm):
    """
    Set the bins and labels for color classification for the selected column.

    parameters:
    -----------

    colm: str
        The data column name for which the bins and labels are to be set.
    """
    method = colm.split("(")[0]
    if "-" in method:
        method = method.split("-")[0]

    thrsholds = set_thresholds(method)
    if not thrsholds:
        return None
    bins = [val[1] for val in thrsholds.values()]
    labels = [key for key in thrsholds.keys()]
    labels.insert(0, labels[0])
    bins.insert(0, next(iter(thrsholds.values()))[0])

    return bins, labels


def plot_anomaly_maps(
    figsize=(25, 20),
    ax_rows=1,
    ax_cols=1,
    df=None,
    x="lon",
    y="lat",
    df_colms=None,
    crs=4326,
    figure_title="",
    figure_title_kwargs={
        "x": 0.5,
        "y": 0.95,
        "fontsize": 15,
        "ha": "center",
        "va": "center",
        "fontweight": "bold",
    },
    maps_titles=None,
    maps_titles_kwargs={"pad": 20, "fontsize": 12, "fontweight": "bold"},
    add_features=True,
    frame_line_width=1,
    cmap="RdYlBu",
    add_gridlines=False,
    cb_kwargs={
        "pos": 0.4,
        "labelsize": 0.5,
        "tick_lines": "center",
        "show_values": False,
    },
):
    m = eomaps.Maps(ax=(ax_rows, ax_cols, 1), figsize=figsize)
    figure_title_kwargs["s"] = figure_title
    m.text(**figure_title_kwargs)

    for i, colm in enumerate(df_colms):
        ax_index = i + 1
        if i == 0:
            m = m
        else:
            m = m.new_map(ax=(ax_rows, ax_cols, ax_index), figsize=(7, 7))
        m.ax.set_title(maps_titles[i], **maps_titles_kwargs)
        m.set_shape.rectangles(radius=0.05)
        if add_features:
            m.add_feature.preset.coastline(lw=0.6)
            m.add_feature.preset.countries(lw=0.4, ls="--")
            m.add_feature.preset.ocean()
            m.add_feature.preset.land()

        m.set_frame(linewidth=frame_line_width)
        m.set_data(data=df, parameter=colm, x=x, y=y, crs=crs)

        if add_gridlines:
            g = m.add_gridlines(
                d=(2, 2),
                ec="grey",
                ls="--",
                lw=0.01,
            )
            g.add_labels(fontsize=8)
        clss = set_bins(colm)
        bins = clss[0] if clss else [0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100]
        labels = (
            clss[1]
            if clss
            else [
                "0-10",
                "10-20",
                "20-30",
                "30-40",
                "40-50",
                "50-60",
                "60-70",
                "70-80",
                "80-90",
                "90-100",
            ]
        )
        vmin = clss[0][0] if clss else df[colm].min()
        vmax = clss[0][-1] if clss else df[colm].max()

        m.set_classify.UserDefined(bins=bins)

        if colm.split("(")[0] == "smds":
            cmap = cmap + "_r"
        m.plot_map(vmin=vmin, vmax=vmax, cmap=cmap, lw=1.5)
        cb = m.add_colorbar(
            label=False, spacing="uniform", pos=0.4, orientation="vertical"
        )
        cb.set_hist_size(0.8)
        cb.tick_params(rotation=0, labelsize=10, pad=5)

        if clss:
            bins, labels = clss
            cb.set_bin_labels(
                bins=bins,
                names=labels,
                tick_lines=cb_kwargs["tick_lines"],
                show_values=cb_kwargs["show_values"],
            )

    m.show()


if __name__ == "__main__":
    import pandas as pd

    df = pd.read_csv("test1.csv")
    df.dropna(inplace=True)

    parameters = [
        "sm-avg(2021-7)",
        "norm-mean(2021-7)",
        "abs-mean(2021-7)",
        "abs-median(2021-7)",
        "zscore(2021-7)",
        "smad(2021-7)",
    ]
    titles = [
        "SSM Average ",
        "Climatology",
        "Absolute Anomaly- Mean",
        "Absolute Anomaly- Median",
        "Z-Score",
        "SMAD",
    ]
    plot_anomaly_maps(
        figsize=(25, 20),
        ax_cols=3,
        ax_rows=2,
        df=df,
        df_colms=parameters,
        figure_title="ASCAT SSM Anomaly Maps for Germany, July 2021",
        maps_titles=titles,
        add_gridlines=False,
        add_features=True,
    )
